#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from sklearn.model_selection import train_test_split


# TODO (minor): enforce versions by runtime checking (sklearn etc)


def split(
    input=["X", "y"],
    output=["Xtr", "Xts", "ytr", "yts"],
    config={"test_size": 0.33, "shuffle": True, "stratify": "y", "random_state": 0},
    **kwargs,
):
    r"""
    >>> from idict import idict, let
    >>> d = idict.fromtoy() >> split
    >>> d.show(colored=False)
    {
        "Xtr": "→(input output config X y)",
        "Xts": "→(input output config X y)",
        "ytr": "→(input output config X y)",
        "yts": "→(input output config X y)",
        "_history": "idict--------------sklearn-1.0.1---split",
        "X": "«{'attr1': {0: 5.1, 1: 1.1, 2: 6.1, 3: 1.1, 4: 3.1, 5: 4.7, 6: 9.1, 7: 8.3, 8: 9.1, 9: 2.5, 10: 7.1, 11: 0.1, 12: 2.1, 13: 0.1, 14: 5.1, 15: 31.1, 16: 1.1, 17: 2.2, 18: 3.1, 19: 1.1}, 'attr2': {0: 6.4, 1: 2.5, 2: 3.6, 3: 3.5, 4: 2.5, 5: 4.9, 6: 3.5, 7: 2.9, 8: 7.2, 9: 4.5, 10: 6.6, 11: 4.3, 12: 0.1, 13: 4.0, 14: 4.5, 15: 4.7, 16: 3.2, 17: 8.5, 18: 2.5, 19: 8.5}}»",
        "y": "«[0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1]»",
        "_id": "8eGk98oY-ZJCTDqlI6cQpv.ScL12.0.1---split",
        "_ids": {
            "Xtr": "XfsV-At5YI8NOBUjjJ-0vEa.qJLAX0I5JQ7SxahM",
            "Xts": "CRh.K81gKNvdsngaWupSWn9JwTpU1O-FiM6QfraE",
            "ytr": "ZzQ4QJL8HWGdMxK5QZl2XLDxq2rVjr4XHPILqMUI",
            "yts": "DojEKdt2iJ0wrptdFUnxUqPGet097oXIBIrZpihN",
            "_history": "c12dMve-KehkFGbdmaqUHg5jXum1pLkGI164PIIC",
            "X": "md_cb8b0c76becc1de32236764e91e8e457e826f (content: 34_1738c83af436029507def2710bc5125f58d0e)",
            "y": "Kp_697ef70c1a293f50ed352ea5775bba6d9b1f4 (content: S0_b6360d62ccafa275d4051dfd02b939104feac)"
        }
    }
    >>> d.yts
    array([1, 0, 1, 0, 0, 0, 1])
    """
    # Multidynamic input is only detected when the kwargs index is also indexed by something.
    multidynamic_input = {}
    for i, _ in enumerate(input):
        multidynamic_input[input[i]] = kwargs[input[i]]

    if "stratify" in config and isinstance(config["stratify"], str):
        if config["stratify"] not in input:  # pragma: no cover
            raise Exception(f"Missing field {config['stratify']} for stratification.")
        config["stratify"] = multidynamic_input[config["stratify"]]
    result = train_test_split(*multidynamic_input.values(), **config)
    out = {k: v for k, v in zip(output, result)}
    # Multidynamic output cannot be detected, so it can only be defined as metadata.
    return out


split.metadata = {
    "id": "idict--------------sklearn-1.0.1---split",
    "name": "split",
    "description": "Split data in two sets.",
    "parameters": ...,
    "code": ...,
    "output": {"fields": [], "auto": ["_history"], "meta": [], "dynamic": ["output"]},
}



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.

import json

from garoupa import Hosh
from garoupa.misc.colors import id2ansi

from ldict.customjson import CustomJSONEncoder
from ldict.lazyval import LazyVal


def idict2txt(d, all, history):
    r"""
    Textual representation of a ldict object

    >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
    >>> from ldict.core.appearance import decolorize
    >>> d = idict(x=1,y=2)
    >>> decolorize(idict2txt(d, False, False))
    '{\n    "x": 1,\n    "y": 2,\n    "_id": "5G_358b45f49c547174eb4bd687079b30cbbe724",\n    "_ids": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 S-_074b5a806933d64f111a93af359a278402f83"\n}'
    >>> decolorize(idict2txt(d, True, False))
    '{\n    "x": 1,\n    "y": 2,\n    "_id": "5G_358b45f49c547174eb4bd687079b30cbbe724",\n    "_ids": {\n        "x": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 (content: l8_09c7059156c4ed2aea46243e9d4b36c01f272)",\n        "y": "S-_074b5a806933d64f111a93af359a278402f83 (content: -B_305c3d0e44c94a5418d982f7dfe8a537a5c4b)"\n    }\n}'

    Parameters
    ----------
    d
    all

    Returns
    -------

    """
    dic = idict2dict(d, all, history)
    txt = json.dumps(dic, indent=4, ensure_ascii=False, cls=CustomJSONEncoder)
    if "_id" in dic:
        txt = txt.replace(dic["_id"], d.hosh.idc)
    if "_history" in dic:
        if isinstance(dic["_history"], LazyVal):
            dic["_history"] = dic["_history"]()
        if isinstance(dic["_history"], str):
            for id in dic["_history"].split(" "):
                txt = txt.replace(id, id2ansi(id))
    if all:
        for k, v in d.hoshes.items():
            nokey = ""
            if k in d.hashes:
                hash = v // k.encode()
                nokey = f" (content: {hash.idc})"
            txt = txt.replace(v.id, v.idc + nokey)  # REMINDER: workaround to avoid json messing with colors
    return txt


def idict2dict(d, all, history):
    # from ldict.core.base import AbstractLazyDict
    dic = d.data.copy()
    if not history and "_history" in dic and isinstance(dic["_history"], dict):
        dic["_history"] = " ".join(Hosh.fromid(k).id for k in dic["_history"])
    if not all:
        if len(d.ids) < 3:
            dic["_ids"] = " ".join(d.ids.values())
        else:
            ids = list(d.ids.values())
            dic["_ids"] = f"{ids[0]}... +{len(d) - 4} ...{ids[-1]}"
    elif "_ids" in dic:
        dic["_ids"] = d.ids.copy()
    return dic



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from typing import Iterator


def new(classin="class_", input="[]", translated_input="[]", output="obj", configin="config=None", version=0, **kwargs):
    r"""
    Instantiate a Python class

    'config' overrides 'input', when conflicting keys are provided.

    >>> from idict import idict, let
    >>> d = idict.fromtoy(output_format="df")
    >>> from sklearn.model_selection import StratifiedKFold
    >>> config = {"n_splits":2, "random_state":0, "shuffle":True}
    >>> d >>= let(new, class_=StratifiedKFold, config=config)
    >>> d.obj
    StratifiedKFold(n_splits=2, random_state=0, shuffle=True)
    """
    if input == "[]":
        input = []
    if translated_input == "[]":
        translated_input = []
    class_ = kwargs[classin]
    config = kwargs[configin]
    if translated_input:
        multidynamic_input = {}
        for i, tr in enumerate(translated_input):
            multidynamic_input[tr] = kwargs[input[i]]
        result = class_(**multidynamic_input, **config)
    else:
        multidynamic_input = [kwargs[_i] for _i in input]
        result = class_(*multidynamic_input, **config)
    return {output: result, "_history": ...}


# REMINDER: A field cannot be used to define a dependence on another field.
#   Only parameters can do that.
#   Otherwise, it would lead to kwargs[kwargs[field]], whose detection is messy at the application step.
def call(
    field="obj",
    methodin="method",
    input="[]",
    translated_input="[]",
    output="call",
    configin="config=None",
    version=0,
    **kwargs
):
    r"""
    Call a method on a field

    Evaluate iterators, if any.
    'config' overrides 'input', when conflicting keys are provided.

    >>> from idict import idict, let
    >>> d = idict.fromtoy(output_format="Xy")
    >>> from sklearn.model_selection import StratifiedKFold
    >>> config = {"n_splits":3, "random_state":0, "shuffle":True}
    >>> d >>= let(new, class_=StratifiedKFold, config=config)
    >>> d.obj
    StratifiedKFold(n_splits=3, random_state=0, shuffle=True)
    >>> d >>= let(call, method="split", input=["X","y"], output=["partition1", "partition2", "partition3"])
    >>> d.partition2
    (array([ 0,  2,  5,  6,  7,  8,  9, 11, 12, 13, 14, 16, 17]), array([ 1,  3,  4, 10, 15, 18, 19]))
    """
    # Multidynamic input is only detected when the kwargs index is also indexed by something.
    if input == "[]":
        input = []
    if translated_input == "[]":
        translated_input = []
    config = kwargs[configin] if configin in kwargs else {}
    if translated_input:
        multidynamic_input = {}
        for i, tr in enumerate(translated_input):
            multidynamic_input[tr] = kwargs[input[i]]
        result = getattr(kwargs[field], kwargs[methodin])(**multidynamic_input, **config)
    else:
        multidynamic_input = [kwargs[_i] for _i in input]
        result = getattr(kwargs[field], kwargs[methodin])(*multidynamic_input, **config)

    if isinstance(result, Iterator):
        result = list(result)
    if isinstance(output, str):
        return {output: result}
    # Multidynamic output cannot be detected, so it can only be defined as metadata.
    return {k: v for k, v in zip(output, result)}


def at(field="obj", indexin="index", output="at", version=0, **kwargs):
    r"""
    Access value inside an indexed field

    Evaluate value-iterator, if it is the case.

    >>> from idict import idict, let
    >>> d = idict.fromtoy(output_format="df")
    >>> from sklearn.model_selection import StratifiedKFold
    >>> config = {"n_splits":3, "random_state":0, "shuffle":True}
    >>> d >>= let(new, class_=StratifiedKFold, config=config)
    >>> d.obj
    StratifiedKFold(n_splits=3, random_state=0, shuffle=True)
    >>> d["y"] = d.df.iloc[:,-1]
    >>> d >>= let(call, method="split", input=["df", "y"], translated_input=["X", "y"], output=[f"p{i}" for i in range(3)])
    >>> d["itr"], d["its"] = d.p2
    >>> d.its
    array([ 2,  6,  9, 11, 12, 17])
    >>> d >>= let(at, field="df", indexin="its", output="ts")
    >>> d.ts
        attr1  attr2  class
    2     6.1    3.6      0
    6     9.1    3.5      0
    9     2.5    4.5      1
    11    0.1    4.3      1
    12    2.1    0.1      0
    17    2.2    8.5      1
    """
    obj = kwargs[field]
    if hasattr(obj, "iloc"):
        result = obj.iloc[kwargs[indexin]]
    else:
        result = obj[kwargs[indexin]]
    if isinstance(result, Iterator):
        result = list(result)
    if isinstance(output, str):
        return {output: result}
    # Multidynamic output cannot be detected, so it can only be defined as metadata.
    return {k: v for k, v in zip(output, result)}


def access(field="obj", propertyin="property", output="access", version=0, **kwargs):
    r"""
    Access a property on a field

    >>> from idict import idict, let
    >>> d = idict.fromtoy(output_format="df")
    >>> cache = {}
    >>> d >>= let(access, field="df", property="head", output="head") >> [cache]
    >>> d.head
    <bound method NDFrame.head of     attr1  attr2  class
    0     5.1    6.4      0
    1     1.1    2.5      1
    2     6.1    3.6      0
    3     1.1    3.5      1
    4     3.1    2.5      0
    5     4.7    4.9      1
    6     9.1    3.5      0
    7     8.3    2.9      1
    8     9.1    7.2      0
    9     2.5    4.5      1
    10    7.1    6.6      0
    11    0.1    4.3      1
    12    2.1    0.1      0
    13    0.1    4.0      1
    14    5.1    4.5      0
    15   31.1    4.7      1
    16    1.1    3.2      0
    17    2.2    8.5      1
    18    3.1    2.5      0
    19    1.1    8.5      1>
    >>> d = idict(d.id, cache)
    >>> d.history
    {'idict--------------------wrapper--access': {'name': 'access', 'description': 'Access a property on a given field.', 'parameters': {'field': 'df', 'propertyin': 'property', 'output': 'head', 'version': 0, 'property': 'head'}, 'code': "def f(field='obj', propertyin='property', output='access', version=0, **kwargs):\nreturn {output: getattr(kwargs[field], kwargs[propertyin]), '_history': ...}"}}
    """
    return {output: getattr(kwargs[field], kwargs[propertyin]), "_history": ...}


def apply(
    functionin="function",
    input="[]",
    translated_input="[]",
    output=["apply"],
    configin="config=None",
    version=0,
    **kwargs
):
    r"""
    >>> from sklearn.utils import resample
    >>> from idict import idict, let
    >>> d = idict(X=[[1,2,3], [4,5,6], [11,12,13]], y=[7,8,9])
    >>> d >>= let(apply, function=resample, input=["X", "y"], output=["X", "y"], config={"n_samples":2, "random_state":0})
    >>> d.X
    [[1, 2, 3], [4, 5, 6]]
    >>> d.y
    [7, 8]
    """
    if input == "[]":
        input = []
    if translated_input == "[]":
        translated_input = []
    f = kwargs[functionin]
    config = kwargs[configin]
    if translated_input:
        multidynamic_input = {}
        for i, tr in enumerate(translated_input):
            multidynamic_input[tr] = kwargs[input[i]]
        result = f(**multidynamic_input, **config)
    else:
        multidynamic_input = [kwargs[_i] for _i in input]
        result = f(*multidynamic_input, **config)

    out = {k: v for k, v in zip(output, result)}
    # Multidynamic output cannot be detected, so it can only be defined as metadata.
    return out


new.metadata = {
    "id": "idict-----------------------wrapper--new",
    "name": "new",
    "description": "Instantiate a Python class.",
    "parameters": ...,
    "code": ...,
}
call.metadata = {
    "id": "idict----------------------wrapper--call",
    "name": "call",
    "description": "Call a method on a given field.",
    "parameters": ...,
    "code": ...,
    "output": {"fields": [], "auto": ["_history"], "meta": [], "dynamic": ["output"]},
}
at.metadata = {
    "id": "idict------------------------wrapper--at",
    "name": "at",
    "description": "Access value inside an indexed field.",
    "parameters": ...,
    "code": ...,
    "output": {"fields": [], "auto": ["_history"], "meta": [], "dynamic": ["output"]},
}
access.metadata = {
    "id": "idict--------------------wrapper--access",
    "name": "access",
    "description": "Access a property on a given field.",
    "parameters": ...,
    "code": ...,
}
apply.metadata = {
    "id": "idict---------------------wrapper--apply",
    "name": "apply",
    "description": "Apply a function.",
    "parameters": ...,
    "code": ...,
    "output": {"fields": [], "auto": ["_history"], "meta": [], "dynamic": ["output"]},
}



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

"""
Functions to be used directly within an idict workflow
"""
from io import StringIO

from idict.function import isplit


def Xy2M(input=["X", "y"], output="M", **kwargs):
    """
    >>> from idict import idict
    >>> from idict.function.dataset import df2Xy
    >>> d = idict.fromtoy(output_format="df")
    >>> d = d >> df2Xy >> Xy2M
    >>> d.M
    array([[ 5.1,  6.4,  0. ],
           [ 1.1,  2.5,  1. ],
           [ 6.1,  3.6,  0. ],
           [ 1.1,  3.5,  1. ],
           [ 3.1,  2.5,  0. ],
           [ 4.7,  4.9,  1. ],
           [ 9.1,  3.5,  0. ],
           [ 8.3,  2.9,  1. ],
           [ 9.1,  7.2,  0. ],
           [ 2.5,  4.5,  1. ],
           [ 7.1,  6.6,  0. ],
           [ 0.1,  4.3,  1. ],
           [ 2.1,  0.1,  0. ],
           [ 0.1,  4. ,  1. ],
           [ 5.1,  4.5,  0. ],
           [31.1,  4.7,  1. ],
           [ 1.1,  3.2,  0. ],
           [ 2.2,  8.5,  1. ],
           [ 3.1,  2.5,  0. ],
           [ 1.1,  8.5,  1. ]])
    """
    import numpy

    return {output: numpy.column_stack((kwargs[input[0]], kwargs[input[1]]))}


def df2Xy(input="df", Xout="X", yout="y", **kwargs):
    """
    >>> from idict import let, idict
    >>> d = idict.fromminiarff()
    >>> d >>= df2Xy
    >>> d.show(colored=False)
    {
        "X": "→(input Xout yout df)",
        "y": "→(input Xout yout df)",
        "_history": "idict-pandas-1.3.4--sklearn-1.0.1--df2Xy",
        "df": "«{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}»",
        "_id": "0GLFP05fQeW7dVFF0m50N8yVTgl-1.0.1--df2Xy",
        "_ids": {
            "X": "hooghJAIuzoXue9tSVrWrU9ek4Rw--J2NP7DnTVR",
            "y": "UldviM8Kn.smh66ZP7X-D9M1k9IQ4M0DmL6B58PJ",
            "_history": "5ItBBxGIMLsBTpdvIGaFHpdkk.2ERtilWaL0O2e.",
            "df": "q3_b71eb05c4be05eba7b6ae5a9245d5dd70b81b (content: 6X_dc8ccea3b2e46f1c78967fae98b692701dc99)"
        }
    }
    >>> d.y
    array([0, 1])
    """
    from sklearn.preprocessing import LabelEncoder

    le = LabelEncoder()
    df = kwargs[input]
    X_ = df.drop(df.columns[[-1]], axis=1)
    y_ = le.fit_transform(df[df.columns[-1]])
    return {Xout: X_, yout: y_, "_history": ...}


def df2arff(input="df", output="arff", **kwargs):
    """
    >>> from idict import let, idict
    >>> d = idict.fromminiarff()
    >>> d >>= let(df2arff, output="a")
    >>> d.show(colored=False)
    {
        "a": "→(input output df)",
        "_history": "idict---------arff2pandas-1.0.1--df2arff",
        "df": "«{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}»",
        "_id": "Ojq9k7ZSbVjwLGlZ7uuqIyhoJPo.p36mAmav2Wul",
        "_ids": {
            "a": "yfX.AOVXE2hlKyGKeGZ3Klpi3LYxm3PpjcjUaLtE",
            "_history": "FbwPhhohM9oJ2RiZe6NOVCGxpc5Z-6jYgymCTa1J",
            "df": "q3_b71eb05c4be05eba7b6ae5a9245d5dd70b81b (content: 6X_dc8ccea3b2e46f1c78967fae98b692701dc99)"
        }
    }
    >>> d.a
    '@RELATION data\\n\\n@ATTRIBUTE attr1 REAL\\n@ATTRIBUTE attr2 REAL\\n@ATTRIBUTE class {0, 1}\\n\\n@DATA\\n5.1,3.5,0\\n3.1,4.5,1\\n'
    """
    from arff2pandas import a2p

    return {output: a2p.dumps(kwargs[input]), "_history": ...}


def openml(Xout="X", yout="y", name="iris", version=1):
    """
    >>> from idict import Ø
    >>> (Ø >> openml).show(colored=False)
    {
        "X": "→(Xout yout name version)",
        "y": "→(Xout yout name version)",
        "_history": "idict--------------sklearn-1.0.1--openml",
        "_id": "idict--------------sklearn-1.0.1--openml",
        "_ids": {
            "X": "3mesqBlZZACO6sy.bZQCc.uyT3vAX0I5JQxOmclE",
            "y": "9.w77PtmNX2vQh25rxwCe6MvK5cU1O-FiMwM4tew",
            "_history": "x5RUSWzvXYmS2WrFfopoDhMo9xC4Zkki-NRROJOc"
        }
    }
    >>> (Ø >> openml).X.head()
       sepallength  sepalwidth  petallength  petalwidth
    0          5.1         3.5          1.4         0.2
    1          4.9         3.0          1.4         0.2
    2          4.7         3.2          1.3         0.2
    3          4.6         3.1          1.5         0.2
    4          5.0         3.6          1.4         0.2
    >>> (Ø >> openml).y.head()
    0    Iris-setosa
    1    Iris-setosa
    2    Iris-setosa
    3    Iris-setosa
    4    Iris-setosa
    Name: class, dtype: category
    Categories (3, object): ['Iris-setosa', 'Iris-versicolor', 'Iris-virginica']
    """
    from sklearn.datasets import fetch_openml

    X, y = fetch_openml(name=name, version=version, as_frame=True, return_X_y=True)
    return {Xout: X, yout: y, "_history": ...}


def arff2df(input="arff", output="df", **kwargs):
    r"""
    >>> from idict import let, idict
    >>> d = idict.fromminiarff(output=["arff"], output_format="arff")
    >>> d.arff
    '@RELATION mini\n@ATTRIBUTE attr1\tREAL\n@ATTRIBUTE attr2 \tREAL\n@ATTRIBUTE class \t{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1'
    >>> d >>= arff2df
    >>> d.show(colored=False)  # doctest:+ELLIPSIS
    {
        "df": "→(input output arff)",
        "_name": "→(input output arff)",
        "_history": "idict---------arff2pandas-1.0.1--arff2df",
        "arff": "@RELATION mini\n@ATTRIBUTE attr1\tREAL\n@ATTRIBUTE attr2 \tREAL\n@ATTRIBUTE class \t{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1",
        "_id": "XraBH1sOCC.9ohqV3hfIRTE5FV3.0.1--arff2df",
        "_ids": {
            "df": "UANHM97QvqLnnzzID7JTDbqRpqUQX-K1K4AEnTby",
            "_name": "gSpa-e2rKoEZwN4N9d5bs83RCbE82M1Cj0zC585q",
            "_history": "ltnWRK7riNIANfskLtRiJOT3rmmQYmhmZxP.N2Yi",
            "arff": "Z._c3e2b235b697e9734b9ec13084129dc30e45b (content: Ev_8bb973161e5ae900c5743b3c332b4a64d1955)"
        }
    }
    >>> d.name
    'mini'
    >>> d.df
       attr1@REAL  attr2@REAL class@{0,1}
    0         5.1         3.5           0
    1         3.1         4.5           1
    """
    from arff2pandas import a2p

    relation = "<Unnamed>"
    with StringIO() as f:
        f.write(kwargs[input])
        text = f.getvalue()
        df = a2p.loads(text)
        for line in isplit(text, "\n"):
            if line[:9].upper() == "@RELATION":
                relation = line[9:].strip()
                break

    return {output: df, "_name": relation, "_history": ...}


df2Xy.metadata = {
    "id": "idict-pandas-1.3.4--sklearn-1.0.1--df2Xy",
    "name": "df2Xy",
    "description": "DataFrame (pandas) to X,y (pandas) converter.",
    "parameters": ...,
    "code": ...,
}
df2arff.metadata = {
    "id": "idict---------arff2pandas-1.0.1--df2arff",
    "name": "df2arff",
    "description": "DataFrame (pandas) to ARFF converter.",
    "parameters": ...,
    "code": ...,
}
openml.metadata = {
    "id": "idict--------------sklearn-1.0.1--openml",
    "name": "openml",
    "description": "Fetch DataFrame+Series (pandas) from OpenML.",
    "parameters": ...,
    "code": ...,
}
arff2df.metadata = {
    "id": "idict---------arff2pandas-1.0.1--arff2df",
    "name": "arff2df",
    "description": "ARFF to DataFrame (pandas) converter.",
    "parameters": ...,
    "code": ...,
}



import json


def value2text(input="value", output="text", **kwargs):
    """
    >>> import numpy as np
    >>> X = [[0, "a", 1.6], [3.2, "b", 2], [8, "c", 3]]
    >>> value2text(value=X)
    {'text': '[[0, "a", 1.6], [3.2, "b", 2], [8, "c", 3]]', '_history': Ellipsis}
    """
    return {output: json.dumps(kwargs[input]), "_history": ...}


value2text.metadata = {
    "id": "------------------------json--value2text",
    "name": "value2text",
    "description": "Generate a JSON-formatted text representing the given object.",
    "parameters": ...,
    "code": ...,
}



from idict.parameter.ifunctionspace import iFunctionSpace

compose = iFunctionSpace.fromfunctions

# TODO (minor): apply(fin="f", input=["X", "y"], output=["z"])     // need detection of multidynamic input (and output?)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from idict import let


def copy(source=None, target=None, **kwargs):
    """
    >>> from idict import idict
    >>> d = idict(x=1) >> let(copy, source="x", target="y")
    >>> d.evaluate()
    >>> d.show(colored=False)
    {
        "y": 1,
        "x": 1,
        "_id": "oRMtnOOH9BxWdFsehRX6pghgYjrwDfaOqWWvzOrq",
        "_ids": {
            "y": "-LB1NvNGUQykT.y2QQGy1E4GbwgwDfaOqWWvzOrq",
            "x": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 (content: l8_09c7059156c4ed2aea46243e9d4b36c01f272)"
        }
    }
    """
    return {target: kwargs[source]}


trcopy = let(copy, source="Xtr", target="X") >> let(copy, source="ytr", target="y")
tscopy = let(copy, source="Xts", target="X") >> let(copy, source="yts", target="y")
# TODO (minor):  finish copy



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
import json

from ldict.core.base import AbstractLazyDict
from ldict.lazyval import LazyVal


# TODO: store metafield even if idict-id is already stored
def storevalue_func(cache):
    def f(k, id, value):
        cache[id] = value

    return f


def storeblob_func(cache, blobs):
    def f(k, id, value):
        if k in blobs:
            cache.setblob(id, blobs[k])
        else:
            cache[id] = value

    return f


def cached(d, cache) -> AbstractLazyDict:
    """
    Store each value (fid: value) and an extra value containing the fids (did: {"_id": did, "_ids": fids}).
    When the dict is a singleton, we have to use "_"+id[1:] as dict id to workaround the ambiguity did=fid.

    Lock the id during the job, to avoid duplicate jobs in a distributed system, if supported by the provided cache.
    """
    # TODO: gravar hashes como aliases no cache pros hoshes. tb recuperar. [serve p/ poupar espaço. e tráfego se usar duplo cache local-remoto]
    #  mas hash não é antecipável! 'cached' teria de fazer o ponteiro: ho -> {"_id": ". . ."}.  aproveitar pack() para guardar todo valor assim.
    from idict.core.idict_ import Idict
    from idict.core.frozenidentifieddict import FrozenIdentifiedDict

    # TODO (minor): do the same for fetch(). useful in the future if needed to speedup syncing of caches avoiding *pack
    store = storeblob_func(cache, d.blobs) if hasattr(cache, "setblob") else storevalue_func(cache)
    front_id = "_" + d.id[1:]

    def closure(outputf, fid, fids, data, output_fields, id):
        def func(**kwargs):
            # Try loading.
            if fid in cache:
                return get_following_pointers(fid, cache)

            # Lock the id for this job.
            if hasattr(cache, "lock"):
                if (t := cache.lockid(fid)) is not None:
                    raise LockedEntryException(f"There is already a job producing the data {fid}, since {t}.")

            # Process and save (all fields, to avoid a parcial idict being stored).
            k = None
            for k, v in fids.items():
                if isinstance(data[k], LazyVal):
                    data[k] = data[k](**kwargs)
                if isinstance(data[k], (FrozenIdentifiedDict, Idict)):
                    cache[v] = {"_id": "_" + data[k].id[1:]}
                    data[k] = cached(data[k], cache)
                elif v not in cache:
                    store(k, v, data[k])
            if (result := data[outputf]) is None:  # pragma: no cover
                if k is None:
                    raise Exception(f"No ids")
                raise Exception(f"Key {k} not in output fields: {output_fields}. ids: {fids.items()}")

            front_id_ = front_id
            if hasattr(cache, "user_hosh"):
                # print("has hosh", d.id)
                if front_id_ not in cache:
                    cache[front_id_] = {"_id": id, "_ids": {k: v for k, v in fids.items() if not k.startswith("_")}}
                front_id_ = (d.id * cache.user_hosh).id
            cache[front_id_] = {"_id": id, "_ids": fids}

            # Unlock id.
            if hasattr(cache, "unlock"):
                cache.unlockid(fid)
            return result

        return func

    data = d.data.copy()
    lazies = False
    output_fields = []
    for field, v in list(data.items()):
        if isinstance(v, LazyVal):
            # REMINDER: Check removed because metafield values recovered from cache are lazy.
            # if field.startswith("_"):  # pragma: no cover
            #     raise Exception("Cannot have a lazy value in a metafield.", field)
            output_fields.append(field)
            lazies = True
            id = d.hashes[field].id if field in d.hashes else d.hoshes[field].id
            deps = {"↑": None}
            deps.update(v.deps)
            lazy = LazyVal(field, closure(field, id, d.ids, d.data, output_fields, d.id), deps, data, None)
            data[field] = lazy

    # Eager saving when there are no lazies.
    if not lazies:
        for k, fid in d.ids.items():
            if fid not in cache:
                if isinstance(data[k], (FrozenIdentifiedDict, Idict)):
                    cache[fid] = {"_id": "_" + data[k].id[1:]}
                    data[k] = cached(data[k], cache)
                else:
                    store(k, fid, data[k])
        if front_id not in cache:
            if hasattr(cache, "user_hosh"):
                cache[front_id] = {"_id": d.id, "_ids": {k: v for k, v in d.ids.items() if not k.startswith("_")}}
                front_id = (d.id * cache.user_hosh).id
            cache[front_id] = {"_id": d.id, "_ids": d.ids}

    return d.clone(data)


def build(id, ids, cache, identity, include_blobs=False):
    """Build an idict from a given identity

    >>> from idict import idict
    >>> a = idict(x=5,z=9)
    >>> b = idict(y=7)
    >>> b["d"] = a
    >>> b >>= [cache := {}]
    >>> print(json.dumps(cache, indent=2))
    {
      "WK_6ba95267cec724067d58b3186ecbcaa4253ad": 7,
      "u9_698c410308e557c005cda07ba00c564152401": {
        "_id": "_._72191dfc2ed7d9ff4c35d514b103ac114161f"
      },
      "GS_cb0fda15eac732cb08351e71fc359058b93bd": 5,
      "N8_524991e7434b2d3444007782c4cd0cd887261": 9,
      "_._72191dfc2ed7d9ff4c35d514b103ac114161f": {
        "_id": "r._72191dfc2ed7d9ff4c35d514b103ac114161f",
        "_ids": {
          "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
          "z": "N8_524991e7434b2d3444007782c4cd0cd887261"
        }
      },
      "_U_ab54a36ac6988bc6722654831fc721f5777ae": {
        "_id": "oU_ab54a36ac6988bc6722654831fc721f5777ae",
        "_ids": {
          "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad",
          "d": "u9_698c410308e557c005cda07ba00c564152401"
        }
      }
    }
    >>> build(b.id, b.ids, cache, b.hosh.ø).evaluated.show(colored=False)
    {
        "y": 7,
        "d": {
            "x": 5,
            "z": 9,
            "_id": "r._72191dfc2ed7d9ff4c35d514b103ac114161f",
            "_ids": {
                "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
                "z": "N8_524991e7434b2d3444007782c4cd0cd887261"
            }
        },
        "_id": "oU_ab54a36ac6988bc6722654831fc721f5777ae",
        "_ids": {
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)",
            "d": "u9_698c410308e557c005cda07ba00c564152401 (content: r._72191dfc2ed7d9ff4c35d514b103ac114161f)"
        }
    }
    >>> (a.hosh ** b"d").show(colored=False)
    u9_698c410308e557c005cda07ba00c564152401
    >>> a = idict(x=5)
    >>> b = idict(y=7)
    >>> b["d"] = a
    >>> b >>= [cache := {}]
    >>> print(json.dumps(cache, indent=2))
    {
      "WK_6ba95267cec724067d58b3186ecbcaa4253ad": 7,
      "I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af": {
        "_id": "_S_cb0fda15eac732cb08351e71fc359058b93bd"
      },
      "GS_cb0fda15eac732cb08351e71fc359058b93bd": 5,
      "_S_cb0fda15eac732cb08351e71fc359058b93bd": {
        "_id": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
        "_ids": {
          "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd"
        }
      },
      "_L_a791c4ff8388e3923e169ce05a0a152def44d": {
        "_id": "DL_a791c4ff8388e3923e169ce05a0a152def44d",
        "_ids": {
          "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad",
          "d": "I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af"
        }
      }
    }
    >>> d = build(b.id, b.ids, cache, b.hosh.ø)
    >>> d.show(colored=False)
    {
        "y": "→(↑)",
        "d": "→(↑)",
        "_id": "DL_a791c4ff8388e3923e169ce05a0a152def44d",
        "_ids": {
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)",
            "d": "I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af (content: GS_cb0fda15eac732cb08351e71fc359058b93bd)"
        }
    }
    >>> d.evaluated.show(colored=False)
    {
        "y": 7,
        "d": {
            "x": 5,
            "_id": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
            "_ids": {
                "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd"
            }
        },
        "_id": "DL_a791c4ff8388e3923e169ce05a0a152def44d",
        "_ids": {
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)",
            "d": "I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af (content: GS_cb0fda15eac732cb08351e71fc359058b93bd)"
        }
    }
    >>> (a.hosh ** b"d").show(colored=False)
    I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af
    >>> a = idict(x=5,z=9)
    >>> b = idict(y=7)
    >>> b["d"] = lambda y: a
    >>> b >>= [cache := {}]
    >>> _ = b.d
    >>> print(json.dumps(cache, indent=2))  # doctest:+ELLIPSIS
    {
      "h8xF5VS3x8yt.ZM3I1D1oKzdS6GcYmec-Y0cnAIl": {
        "_id": "_._72191dfc2ed7d9ff4c35d514b103ac114161f"
      },
      "GS_cb0fda15eac732cb08351e71fc359058b93bd": 5,
      "N8_524991e7434b2d3444007782c4cd0cd887261": 9,
      "_._72191dfc2ed7d9ff4c35d514b103ac114161f": {
        "_id": "r._72191dfc2ed7d9ff4c35d514b103ac114161f",
        "_ids": {
          "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
          "z": "N8_524991e7434b2d3444007782c4cd0cd887261"
        }
      },
      "WK_6ba95267cec724067d58b3186ecbcaa4253ad": 7,
      "_P9idohZcsojDgKThROI-YoSzRBcYmec-Y0cnAIl": {
        "_id": "UP9idohZcsojDgKThROI-YoSzRBcYmec-Y0cnAIl",
        "_ids": {
          "d": "h8xF5VS3x8yt.ZM3I1D1oKzdS6GcYmec-Y0cnAIl",
          "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad"
        }
      }
    }
    >>> build(b.id, b.ids, cache, b.hosh.ø).evaluated.show(colored=False)  # doctest:+ELLIPSIS
    {
        "d": {
            "x": 5,
            "z": 9,
            "_id": "r._72191dfc2ed7d9ff4c35d514b103ac114161f",
            "_ids": {
                "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
                "z": "N8_524991e7434b2d3444007782c4cd0cd887261"
            }
        },
        "y": 7,
        "_id": "...",
        "_ids": {
            "d": "...",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> (a.hosh ** b"d").show(colored=False)
    u9_698c410308e557c005cda07ba00c564152401
    >>> a = idict(x=5)
    >>> b = idict(y=7)
    >>> b["d"] = lambda y: a
    >>> b >>= [cache := {}]
    >>> _ = b.d
    >>> print(json.dumps(cache, indent=2))  # doctest:+ELLIPSIS
    {
      "...": {
        "_id": "_S_cb0fda15eac732cb08351e71fc359058b93bd"
      },
      "GS_cb0fda15eac732cb08351e71fc359058b93bd": 5,
      "_S_cb0fda15eac732cb08351e71fc359058b93bd": {
        "_id": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
        "_ids": {
          "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd"
        }
      },
      "WK_6ba95267cec724067d58b3186ecbcaa4253ad": 7,
      "...": {
        "_id": "...",
        "_ids": {
          "d": "...",
          "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad"
        }
      }
    }
    >>> d = build(b.id, b.ids, cache, b.hosh.ø)
    >>> d.show(colored=False)  # doctest:+ELLIPSIS
    {
        "d": "→(↑)",
        "y": "→(↑)",
        "_id": "...",
        "_ids": {
            "d": "...",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> d.evaluated.show(colored=False)  # doctest:+ELLIPSIS
    {
        "d": {
            "x": 5,
            "_id": "GS_cb0fda15eac732cb08351e71fc359058b93bd",
            "_ids": {
                "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd"
            }
        },
        "y": 7,
        "_id": "...",
        "_ids": {
            "d": "...",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> (a.hosh ** b"d").show(colored=False)
    I0_dac96298c4c5bf8cb0cde8d8eb3e4a78ca1af
    """
    from idict.core.frozenidentifieddict import FrozenIdentifiedDict

    if include_blobs:
        raise NotImplementedError
    hosh = identity * id
    data, hashes, hoshes = {}, {}, {}
    for k, fid in ids.items():
        # REMINDER: An item id will never start with '_'. That only happens with singleton-idict id translated to cache.
        if fid in cache:
            value = get_following_pointers(fid, cache)
            # WARN: The closures bellow assume items will not be removed from 'cache' in the meantime.
            if isinstance(value, dict) and list(value.keys()) == ["_id", "_ids"]:
                closure = lambda value_: lambda **kwargs: build(value_["_id"], value_["_ids"], cache, identity)
                data[k] = LazyVal(k, closure(value), {"↑": None}, {}, None)
            else:
                closure = lambda fid_: lambda **kwargs: cache[fid_]
                data[k] = LazyVal(k, closure(fid), {"↑": None}, {}, None)
        else:  # pragma: no cover
            raise Exception(f"Missing key={fid} or singleton key=_{fid[1:]}.\n{json.dumps(cache, indent=2)}")
        hoshes[k] = identity * fid
        if fid[2] == "_":
            hashes[k] = hoshes[k] // k.encode()

    internals = dict(blobs={}, hashes=hashes, hoshes=hoshes, hosh=hosh)
    return FrozenIdentifiedDict(data, identity=identity, _cloned=internals)


def get_following_pointers(fid, cache):
    """Fetch item value from cache following pointers"""
    result = cache[fid]
    while isinstance(result, dict) and list(result.keys()) == ["_id"]:
        result = cache[result["_id"]]
    return result


class LockedEntryException(Exception):
    pass
